from ast import *
from copy import copy

"""
Rewrites all occurences of augmented assignments
into normal assignments.
"""


class RewriteAugAssign(NodeTransformer):
    def visit_AugAssign(self, node: AugAssign) -> Assign:
        target_cp = copy(node.target)
        target_cp.ctx = Load()
        a = Assign(
            [self.visit(node.target)],
            BinOp(
                self.visit(target_cp),
                self.visit(node.op),
                self.visit(node.value),
            ),
        )
        return a



import ast

from ast import *
from dataclasses import dataclass
import typing


class Type:
    pass


@dataclass(unsafe_hash=True)
class InstanceType(Type):
    pass


@dataclass(unsafe_hash=True)
class RecordInstanceType(InstanceType):
    typ: str


@dataclass(unsafe_hash=True)
class UnionInstanceType(InstanceType):
    typs: typing.List[InstanceType]


@dataclass(unsafe_hash=True)
class OptionalInstanceType(InstanceType):
    typ: InstanceType


IntegerType = RecordInstanceType(int.__name__)
StringType = RecordInstanceType(str.__name__)
ByteStringType = RecordInstanceType(bytes.__name__)
BoolType = RecordInstanceType(bool.__name__)
UnitType = RecordInstanceType("Unit")


@dataclass(frozen=True, unsafe_hash=True)
class Record:
    name: str
    constructor: int
    fields: typing.List[typing.Tuple[str, Type]]


@dataclass(unsafe_hash=True)
class ClassType(Type):
    pass


@dataclass(unsafe_hash=True)
class RecordType(ClassType):
    record: Record


@dataclass(unsafe_hash=True)
class UnionType(ClassType):
    typs: typing.List[ClassType]


@dataclass(unsafe_hash=True)
class OptionalType(ClassType):
    typ: ClassType


NoneRecord = Record("None", 0, [])
NoneInstanceType = RecordInstanceType("None")
NoneType = RecordType(NoneRecord)


@dataclass(unsafe_hash=True)
class TupleType(Type):
    typs: typing.List[Type]


@dataclass(unsafe_hash=True)
class ListType(Type):
    typ: Type


@dataclass(unsafe_hash=True)
class DictType(Type):
    key_typ: Type
    value_typ: Type


@dataclass(unsafe_hash=True)
class FunctionType(Type):
    argtyps: typing.List[Type]
    rettyp: Type


class TypedAST(AST):
    typ: Type


class typedexpr(TypedAST, expr):
    pass


class typedstmt(TypedAST, stmt):
    # Statements always have type None
    typ = NoneType


class typedarg(TypedAST, arg):
    pass


class typedarguments(TypedAST, arguments):
    args: typing.List[typedarg]
    vararg: typing.Union[typedarg, None]
    kwonlyargs: typing.List[typedarg]
    kw_defaults: typing.List[typing.Union[typedexpr, None]]
    kwarg: typing.Union[typedarg, None]
    defaults: typing.List[typedexpr]


class TypedModule(typedstmt, Module):
    body: typing.List[typedstmt]


class TypedFunctionDef(typedstmt, FunctionDef):
    body: typing.List[typedstmt]
    args: arguments


class TypedIf(typedstmt, If):
    test: typedexpr
    body: typing.List[typedstmt]
    orelse: typing.List[typedstmt]


class TypedReturn(typedstmt, Return):
    value: typedexpr


class TypedExpression(typedexpr, Expression):
    body: typedexpr


class TypedCall(typedexpr, Call):
    func: typedexpr
    args: typing.List[typedexpr]


class TypedExpr(typedstmt, Expr):
    value: typedexpr


class TypedAssign(typedstmt, Assign):
    targets: typing.List[typedexpr]
    value: typedexpr


class TypedWhile(typedstmt, While):
    test: typedexpr
    body: typing.List[typedstmt]
    orelse: typing.List[typedstmt]


class TypedFor(typedstmt, For):
    target: typedexpr
    iter: typedexpr
    body: typing.List[typedstmt]
    orelse: typing.List[typedstmt]


class TypedPass(typedstmt, Pass):
    pass


class TypedName(typedexpr, Name):
    pass


class TypedConstant(TypedAST, Constant):
    pass


class TypedTuple(typedexpr, Tuple):
    typ: typing.List[TypedAST]


class TypedList(typedexpr, List):
    typ: typing.List[TypedAST]


class TypedCompare(typedexpr, Compare):
    left: typedexpr
    ops: typing.List[cmpop]
    comparators: typing.List[typedexpr]


class TypedBinOp(typedexpr, BinOp):
    left: typedexpr
    right: typedexpr


class TypedUnaryOp(typedexpr, UnaryOp):
    operand: typedexpr


class TypedSubscript(typedexpr, Subscript):
    value: typedexpr


class TypedAttribute(typedexpr, Attribute):
    value: typedexpr
    pos: int


class TypeInferenceError(AssertionError):
    pass


def type_from_annotation(ann: expr):
    if isinstance(ann, Constant):
        if ann.value is None:
            return NoneType
    if isinstance(ann, Tuple):
        if not ann.elts:
            # This is ()
            return UnitType
    if isinstance(ann, Name):
        return RecordInstanceType(ann.id)
    if isinstance(ann, Subscript):
        assert isinstance(
            ann.value, Name
        ), "Only Optional, Union, Dict and List are allowed as Generic types"
        assert isinstance(ann.slice, Index), "Generic types must be parameterized"
        if ann.value.id == "Optional":
            ann_type = type_from_annotation(ann.slice.value)
            assert isinstance(
                ann_type, InstanceType
            ), "Optional must have a single type as parameter"
            return OptionalInstanceType(ann_type)
        if ann.value.id == "Union":
            assert isinstance(
                ann.slice.value, Tuple
            ), "Union must combine multiple classes"
            ann_types = [type_from_annotation(e) for e in ann.slice.value.elts]
            assert all(
                isinstance(e, InstanceType) for e in ann_types
            ), "Union must combine multiple classes"
            return UnionInstanceType(ann_types)
        if ann.value.id == "List":
            return ListType(type_from_annotation(ann.slice.value))
        if ann.value.id == "Dict":
            assert isinstance(ann.slice.value, Tuple), "Dict must combine two classes"
            assert len(ann.slice.value.elts) == 2, "Dict must combine two classes"
            return DictType(
                type_from_annotation(ann.slice.value.elts[0]),
                type_from_annotation(ann.slice.value.elts[1]),
            )
        raise NotImplementedError(
            "Only Optional, Union, Dict and List are allowed as Generic types"
        )
    if ann is None:
        TypeInferenceError(
            "Type annotation is missing for a function argument or return value"
        )
    raise NotImplementedError(f"Annotation type {ann} is not supported")


class RecordReader(NodeVisitor):
    name: str
    constructor: int
    attributes: typing.List[typing.Tuple[str, Type]]

    def __init__(self):
        self.constructor = 0
        self.attributes = []

    @classmethod
    def extract(cls, c: ClassDef) -> Record:
        f = cls()
        f.visit(c)
        return Record(f.name, f.constructor, f.attributes)

    def visit_AnnAssign(self, node: AnnAssign) -> None:
        assert isinstance(
            node.target, Name
        ), "Record elements must have named attributes"
        if node.target.id != "CONSTR_ID":
            assert (
                node.value is None
            ), f"PlutusData attribute {node.target.id} may not have a default value"
            self.attributes.append(
                (node.target.id, type_from_annotation(node.annotation))
            )
            return
        assert isinstance(
            node.value, Constant
        ), "CONSTR_ID must be assigned a constant integer"
        assert isinstance(
            node.value.value, int
        ), "CONSTR_ID must be assigned an integer"
        self.constructor = node.value.value

    def visit_ClassDef(self, node: ClassDef) -> None:
        self.name = node.name
        for s in node.body:
            self.visit(s)

    def visit_Pass(self, node: Pass) -> None:
        pass

    def visit_Assign(self, node: Assign) -> None:
        assert len(node.targets) == 1, "Record elements must be assigned one by one"
        target = node.targets[0]
        assert isinstance(target, Name), "Record elements must have named attributes"
        assert (
            target.id == "CONSTR_ID"
        ), "Type annotations may only be omitted for CONSTRUCTOR"
        assert isinstance(
            node.value, Constant
        ), "CONSTR_ID must be assigned a constant integer"
        assert isinstance(
            node.value.value, int
        ), "CONSTR_ID must be assigned an integer"
        self.constructor = node.value.value

    def visit_Expr(self, node: Expr) -> None:
        assert isinstance(
            node.value, Constant
        ), "Only comments are allowed inside classes"
        return None

    def generic_visit(self, node: AST) -> None:
        raise NotImplementedError(f"Can not compile {node} inside of a class")



from enum import Enum
from ast import parse

from .type_inference import *
from .rewrite_tuple_assign import RewriteTupleAssign
from .rewrite_augassign import RewriteAugAssign
from .rewrite_import_plutusdata import RewriteImportPlutusData
from .rewrite_import_typing import RewriteImportTyping
from .rewrite_import import RewriteImport

import pluthon as plt

STATEMONAD = "s"


BinOpMap = {
    Add: {
        IntegerType: {
            IntegerType: plt.AddInteger,
        },
        ByteStringType: {
            ByteStringType: plt.AppendByteString,
        },
    },
    Sub: {
        IntegerType: {
            IntegerType: plt.SubtractInteger,
        }
    },
    Mult: {
        IntegerType: {
            IntegerType: plt.MultiplyInteger,
        }
    },
    Div: {
        IntegerType: {
            IntegerType: plt.DivideInteger,
        }
    },
    Mod: {
        IntegerType: {
            IntegerType: plt.ModInteger,
        }
    },
}

CmpMap = {
    Eq: {
        IntegerType: {
            IntegerType: plt.EqualsInteger,
        },
        ByteStringType: {
            ByteStringType: plt.EqualsByteString,
        },
        BoolType: {
            BoolType: plt.EqualsBool,
        },
    },
    Lt: {
        IntegerType: {
            IntegerType: plt.LessThanInteger,
        },
    },
}

TransformExtParamsMap = {
    IntegerType: lambda x: plt.UnIData(x),
    StringType: lambda x: plt.DecodeUtf8(plt.UnBData(x)),
    ByteStringType: lambda x: plt.UnBData(x),
    ListType: lambda x: plt.UnListData(x),
    DictType: lambda x: plt.UnMapData(x),
    UnitType: lambda x: plt.Lambda(["_"], plt.Unit()),
    BoolType: lambda x: plt.NotEqualsInteger(plt.UnIData(x), plt.Integer(0)),
}
TransformOutputMap = {
    IntegerType: lambda x: plt.IData(x),
    StringType: lambda x: plt.BData(plt.EncodeUtf8(x)),
    ByteStringType: lambda x: plt.BData(x),
    ListType: lambda x: plt.ListData(x),
    DictType: lambda x: plt.MapData(x),
    UnitType: lambda x: plt.Lambda(["_"], plt.Unit()),
    BoolType: lambda x: plt.Ite(x, plt.Unit(), plt.TraceError("ValidationError")),
}

ConstantMap = {
    str: plt.Text,
    bytes: plt.ByteString,
    int: plt.Integer,
    bool: plt.Bool,
    type(None): lambda _: plt.NoneData(),
}


def extend_statemonad(
    names: typing.List[str],
    values: typing.List[plt.AST],
    old_statemonad: plt.FunctionalMap,
):
    return plt.FunctionalMapExtend(old_statemonad, [n.encode() for n in names], values)


class PythonBuiltIn(Enum):
    print = plt.Lambda(
        ["f", "x", STATEMONAD],
        plt.Trace(plt.Var("x"), plt.NoneData()),
    )
    range = plt.Lambda(["f", "limit", STATEMONAD], plt.Range(plt.Var("limit")))


INITIAL_STATE = plt.FunctionalMapExtend(
    plt.FunctionalMap(),
    [b.name for b in PythonBuiltIn],
    [b.value for b in PythonBuiltIn],
)


class UPLCCompiler(NodeTransformer):
    """
    Expects a TypedAST and returns UPLC/Pluto like code
    """

    def visit(self, node):
        """Visit a node."""
        node_class_name = node.__class__.__name__
        if node_class_name.startswith("Typed"):
            node_class_name = node_class_name[len("Typed") :]
        method = "visit_" + node_class_name
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def visit_sequence(self, node_seq: typing.List[typedstmt]) -> plt.AST:
        s = plt.Var(STATEMONAD)
        for n in node_seq:
            compiled_stmt = self.visit(n)
            s = plt.Apply(compiled_stmt, s)
        return plt.Lambda([STATEMONAD], s)

    def visit_BinOp(self, node: TypedBinOp) -> plt.AST:
        opmap = BinOpMap.get(type(node.op))
        if opmap is None:
            raise NotImplementedError(f"Operation {node.op} is not implemented")
        opmap2 = opmap.get(node.left.typ)
        if opmap2 is None:
            raise NotImplementedError(
                f"Operation {node.op} is not implemented for left type {node.left.typ}"
            )
        op = opmap2.get(node.right.typ)
        if opmap2 is None:
            raise NotImplementedError(
                f"Operation {node.op} is not implemented for left type {node.left.typ} and right type {node.right.typ}"
            )
        return plt.Lambda(
            [STATEMONAD],
            op(
                plt.Apply(self.visit(node.left), plt.Var(STATEMONAD)),
                plt.Apply(self.visit(node.right), plt.Var(STATEMONAD)),
            ),
        )

    def visit_Compare(self, node: Compare) -> plt.AST:
        assert len(node.ops) == 1, "Only single comparisons are supported"
        assert len(node.comparators) == 1, "Only single comparisons are supported"
        opmap = CmpMap.get(type(node.ops[0]))
        if opmap is None:
            raise NotImplementedError(f"Operation {node.ops[0]} is not implemented")
        opmap2 = opmap.get(node.left.typ)
        if opmap2 is None:
            raise NotImplementedError(
                f"Operation {node.ops[0]} is not implemented for left type {node.left.typ}"
            )
        op = opmap2.get(node.comparators[0].typ)
        if op is None:
            raise NotImplementedError(
                f"Operation {node.ops[0]} is not implemented for left type {node.left.typ} and right type {node.comparators[0].typ}"
            )
        return plt.Lambda(
            [STATEMONAD],
            op(
                plt.Apply(self.visit(node.left), plt.Var(STATEMONAD)),
                plt.Apply(self.visit(node.comparators[0]), plt.Var(STATEMONAD)),
            ),
        )

    def visit_Module(self, node: TypedModule) -> plt.AST:
        # find main function
        # TODO can use more sophisiticated procedure here i.e. functions marked by comment
        main_fun = None
        for s in node.body:
            if isinstance(s, FunctionDef) and s.name == "validator":
                main_fun = s
        cp = plt.Program(
            "0.0.1",
            # TODO directly unwrap supposedly int/byte data? how?
            plt.Lambda(
                [f"p{i}" for i, _ in enumerate(main_fun.args.args)],
                TransformOutputMap.get(main_fun.typ.rettyp, lambda x: x)(
                    plt.Let(
                        [
                            ("s", INITIAL_STATE),
                            (
                                "g",
                                plt.FunctionalMapAccess(
                                    plt.Apply(
                                        self.visit_sequence(node.body), plt.Var("s")
                                    ),
                                    plt.ByteString("validator".encode("utf8")),
                                    plt.TraceError("NameError"),
                                ),
                            ),
                        ],
                        plt.Apply(
                            plt.Var("g"),
                            plt.Var("g"),
                            *[
                                TransformExtParamsMap.get(a.typ, lambda x: x)(
                                    plt.Var(f"p{i}")
                                )
                                for i, a in enumerate(main_fun.args.args)
                            ],
                            plt.Var("s"),
                        ),
                    ),
                ),
            ),
        )
        return cp

    def visit_Constant(self, node: TypedConstant) -> plt.AST:
        plt_type = ConstantMap.get(type(node.value))
        if plt_type is None:
            raise NotImplementedError(
                f"Constants of type {type(node.value)} are not supported"
            )
        return plt.Lambda([STATEMONAD], plt_type(node.value))

    def visit_NoneType(self, _: typing.Optional[typing.Any]) -> plt.AST:
        return plt.Lambda([STATEMONAD], plt.NoneData())

    def visit_Assign(self, node: TypedAssign) -> plt.AST:
        assert (
            len(node.targets) == 1
        ), "Assignments to more than one variable not supported yet"
        assert isinstance(
            node.targets[0], Name
        ), "Assignments to other things then names are not supported"
        if isinstance(node.targets[0].typ, ClassType):
            # Assigning a class type to another class type is equivalent to a ClassDef - a nop
            return self.visit_sequence([])
        compiled_e = self.visit(node.value)
        # (\{STATEMONAD} -> (\x -> if (x ==b {self.visit(node.targets[0])}) then ({compiled_e} {STATEMONAD}) else ({STATEMONAD} x)))
        return plt.Lambda(
            [STATEMONAD],
            plt.FunctionalMapExtend(
                plt.Var(STATEMONAD),
                [node.targets[0].id],
                [plt.Apply(compiled_e, plt.Var(STATEMONAD))],
            ),
        )

    def visit_Name(self, node: TypedName) -> plt.AST:
        # depending on load or store context, return the value of the variable or its name
        if isinstance(node.ctx, Load):
            return plt.Lambda(
                [STATEMONAD],
                plt.FunctionalMapAccess(
                    plt.Var(STATEMONAD),
                    plt.ByteString(node.id.encode()),
                    plt.TraceError("NameError"),
                ),
            )
        raise NotImplementedError(f"Context {node.ctx} not supported")

    def visit_Expr(self, node: TypedExpr) -> plt.AST:
        # we exploit UPLCs eager evaluation here
        # the expression is computed even though its value is eventually discarded
        # Note this really only makes sense for Trace
        return plt.Lambda(
            [STATEMONAD],
            plt.Apply(
                plt.Lambda(["_"], plt.Var(STATEMONAD)),
                plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
            ),
        )

    def visit_Call(self, node: TypedCall) -> plt.AST:
        # compiled_args = " ".join(f"({self.visit(a)} {STATEMONAD})" for a in node.args)
        # return rf"(\{STATEMONAD} -> ({self.visit(node.func)} {compiled_args})"
        return plt.Lambda(
            [STATEMONAD],
            plt.Let(
                [("g", plt.Apply(self.visit(node.func), plt.Var(STATEMONAD)))],
                plt.Apply(
                    plt.Var("g"),
                    # pass the function to itself for recursion
                    plt.Var("g"),
                    # pass in all arguments evaluated with the statemonad
                    *(plt.Apply(self.visit(a), plt.Var(STATEMONAD)) for a in node.args),
                    # eventually pass in the state monad as well
                    plt.Var(STATEMONAD),
                ),
            ),
        )

    def visit_FunctionDef(self, node: TypedFunctionDef) -> plt.AST:
        body = node.body.copy()
        if not isinstance(body[-1], Return):
            tr = Return(None)
            tr.typ = NoneType
            assert (
                node.typ.rettyp == NoneType
            ), "Function has no return statement but is supposed to return not-None value"
            body.append(tr)
        compiled_body = self.visit_sequence(body[:-1])
        compiled_return = self.visit(body[-1].value)
        args_state = plt.FunctionalMapExtend(
            plt.Var(STATEMONAD),
            # under the name of the function, it can access itself
            [node.name] + [a.arg for a in node.args.args],
            [plt.Var("f")] + [plt.Var(f"p{i}") for i in range(len(node.args.args))],
        )
        return plt.Lambda(
            [STATEMONAD],
            plt.FunctionalMapExtend(
                plt.Var(STATEMONAD),
                [node.name],
                [
                    plt.Lambda(
                        # expect the statemonad again -> this is the basis for internally available values
                        ["f"]
                        + [f"p{i}" for i in range(len(node.args.args))]
                        + [STATEMONAD],
                        plt.Apply(
                            compiled_return,
                            plt.Apply(
                                compiled_body,
                                args_state,
                            ),
                        ),
                    )
                ],
            ),
        )

    def visit_While(self, node: TypedWhile) -> plt.AST:
        compiled_c = self.visit(node.test)
        compiled_s = self.visit_sequence(node.body)
        if node.orelse:
            # If there is orelse, transform it to an appended sequence (TODO check if this is correct)
            cn = copy(node)
            cn.orelse = []
            return self.visit_sequence([cn] + node.orelse)
        # return rf"(\{STATEMONAD} -> let g = (\s f -> if ({compiled_c} s) then f ({compiled_s} s) f else s) in (g {STATEMONAD} g))"
        return plt.Lambda(
            [STATEMONAD],
            plt.Let(
                bindings=[
                    (
                        "g",
                        plt.Lambda(
                            ["s", "f"],
                            plt.Ite(
                                plt.Apply(compiled_c, plt.Var("s")),
                                plt.Apply(
                                    plt.Var("f"),
                                    plt.Apply(compiled_s, plt.Var("s")),
                                    plt.Var("f"),
                                ),
                                plt.Var("s"),
                            ),
                        ),
                    ),
                ],
                term=plt.Apply(plt.Var("g"), plt.Var(STATEMONAD), plt.Var("g")),
            ),
        )

    def visit_For(self, node: TypedFor) -> plt.AST:
        if node.orelse:
            # If there is orelse, transform it to an appended sequence (TODO check if this is correct)
            cn = copy(node)
            cn.orelse = []
            return self.visit_sequence([cn] + node.orelse)
        if isinstance(node.iter.typ, ListType):
            assert isinstance(
                node.target, Name
            ), "Can only assign value to singleton element"
            return plt.Lambda(
                [STATEMONAD],
                plt.FoldList(
                    plt.Apply(self.visit(node.iter), plt.Var(STATEMONAD)),
                    plt.Lambda(
                        [STATEMONAD, "e"],
                        plt.Apply(
                            self.visit_sequence(node.body),
                            plt.FunctionalMapExtend(
                                plt.Var(STATEMONAD),
                                [node.target.id],
                                [plt.Var("e")],
                            ),
                        ),
                    ),
                    plt.Var(STATEMONAD),
                ),
            )
        raise NotImplementedError(
            "Compilation of for statements for anything but lists not implemented yet"
        )

    def visit_If(self, node: TypedIf) -> plt.AST:
        return plt.Lambda(
            [STATEMONAD],
            plt.Ite(
                plt.Apply(self.visit(node.test), plt.Var(STATEMONAD)),
                plt.Apply(self.visit_sequence(node.body), plt.Var(STATEMONAD)),
                plt.Apply(self.visit_sequence(node.orelse), plt.Var(STATEMONAD)),
            ),
        )

    def visit_Return(self, node: TypedReturn) -> plt.AST:
        raise NotImplementedError(
            "Compilation of return statements except for last statement in function is not supported."
        )

    def visit_Pass(self, node: TypedPass) -> plt.AST:
        return self.visit_sequence([])

    def visit_Subscript(self, node: TypedSubscript) -> plt.AST:
        assert isinstance(
            node.slice, Index
        ), "Only single index slices are currently supported"
        if isinstance(node.value.typ, TupleType):
            assert isinstance(
                node.slice.value, Constant
            ), "Only constant index access for tuples is supported"
            assert isinstance(node.ctx, Load), "Tuples are read-only"
            return plt.Lambda(
                [STATEMONAD],
                plt.FunctionalTupleAccess(
                    plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
                    node.slice.value.value,
                    len(node.value.typ.typs),
                ),
            )
        if isinstance(node.value.typ, ListType):
            # rf'(\{STATEMONAD} -> let g = (\i xs f -> if (!NullList xs) then (!Trace "OOB" (Error ())) else (if i ==i 0 then (!HeadList xs) else f (i -i 1) (!TailList xs) f)) in (g ({compiled_i} {STATEMONAD}) ({compiled_v} {STATEMONAD}) g))'
            return plt.Lambda(
                [STATEMONAD],
                plt.Let(
                    [
                        (
                            "g",
                            plt.Lambda(
                                ["i", "xs", "f"],
                                plt.Ite(
                                    plt.NullList(plt.Var("xs")),
                                    self.visit(
                                        TypedCall(
                                            TypedName("print", ctx=Load()),
                                            [
                                                TypedConstant(
                                                    "Out of bounds",
                                                    RecordInstanceType("str"),
                                                )
                                            ],
                                        )
                                    ),
                                    plt.Ite(
                                        plt.EqualsInteger(
                                            plt.Var("i"),
                                            plt.Integer(0),
                                        ),
                                        plt.HeadList(
                                            plt.Var("xs"),
                                        ),
                                        plt.Apply(
                                            plt.Var("f"),
                                            plt.SubtractInteger(
                                                plt.Var("i"),
                                                plt.Integer(1),
                                            ),
                                            plt.TailList(
                                                plt.Var("xs"),
                                            ),
                                            plt.Var("f"),
                                        ),
                                    ),
                                ),
                            ),
                        )
                    ],
                    plt.Apply(
                        plt.Var("g"),
                        plt.Apply(self.visit(node.slice.value), plt.Var(STATEMONAD)),
                        plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
                        plt.Var("g"),
                    ),
                ),
            )
        raise NotImplementedError(f"Could not implement subscript of {node}")

    def visit_Tuple(self, node: TypedTuple) -> plt.AST:
        return plt.Lambda(
            [STATEMONAD],
            plt.FunctionalTuple(
                *(plt.Apply(self.visit(e), plt.Var(STATEMONAD)) for e in node.elts)
            ),
        )

    def visit_ClassDef(self, node: ClassDef) -> plt.AST:
        # TODO add initializer to state monad
        return self.visit_sequence([])

    def visit_Attribute(self, node: TypedAttribute) -> plt.AST:
        # TODO rewrite to access the field at position node.pos, directly in pluthon
        # (use the internal function for fields)
        # TODO cover case where constr should be accessed
        return self.visit(
            TypedSubscript(
                value=TypedCall(
                    TypedName(
                        id="__fields__",
                        ctx=Load(),
                        typ=FunctionType([RecordInstanceType], ListType(node.typ)),
                    ),
                    [node.value],
                    typ=ListType(node.typ),
                ),
                slice=Index(value=TypedConstant(value=node.pos, typ=IntegerType)),
                typ=node.typ,
            )
        )

    def generic_visit(self, node: AST) -> plt.AST:
        raise NotImplementedError(f"Can not compile {node}")


def compile(prog: AST):
    compiler_steps = [
        # Important to call this one first - it imports all further files
        RewriteImport,
        # The remaining order is almost arbitrary
        RewriteAugAssign,
        RewriteTupleAssign,
        RewriteImportPlutusData,
        RewriteImportTyping,
        AggressiveTypeInferencer,
        UPLCCompiler,
    ]
    for s in compiler_steps:
        prog = s().visit(prog)
    return prog



from ast import *
from copy import copy
from typing import Optional

"""
Checks that there was an import of dataclass if there are any class definitions
"""


class RewriteImportTyping(NodeTransformer):

    imports_typing = False

    def visit_ImportFrom(self, node: ImportFrom) -> Optional[ImportFrom]:
        if node.module != "typing":
            return node
        assert (
            len(node.names) == 4
        ), "The program must contain one 'from typing import List, Dict, Optional, Union'"
        for i, n in enumerate(["List", "Dict", "Optional", "Union"]):
            assert (
                node.names[i].name == n
            ), "The program must contain one 'from typing import List, Dict, Optional, Union'"
            assert (
                node.names[i].asname == None
            ), "The program must contain one 'from typing import List, Dict, Optional, Union'"
        self.imports_typing = True
        return None

    def visit_ClassDef(self, node: ClassDef) -> ClassDef:
        assert (
            self.imports_typing
        ), "typing must be imported in order to use datum classes"
        return node



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

import eopsin

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="eopsin-lang",
    version=eopsin.__version__,
    description="A simple pythonic programming language for Smart Contracts on Cardano ",
    author=eopsin.__author__,
    author_email=eopsin.__author_email__,
    url=eopsin.__url__,
    py_modules=["eopsin"],
    packages=find_packages(),
    install_requires=["pluthon", "pycardano==0.7.2"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=eopsin.__license__,
    classifiers=[
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Compilers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="python cardano smart contract blockchain verification haskell",
    python_requires=">=3",
)



from copy import copy
from ast import *

# from frozendict import frozendict

from .typed_ast import *

"""
An aggressive type inference based on the work of Aycock [1].
It only allows a subset of legal python operations which
allow us to infer the type of all involved variables
statically.
Using this we can resolve overloaded functions when translating Python
into UPLC where there is no dynamic type checking.
Additionally, this conveniently implements an additional layer of
security into the Smart Contract by checking type correctness.


[1]: https://legacy.python.org/workshops/2000-01/proceedings/papers/aycock/aycock.html
"""


INITIAL_SCOPE = dict(
    {
        "print": FunctionType([StringType], NoneType),
        "range": FunctionType(
            [IntegerType],
            ListType(IntegerType),
        ),
    }
)


class AggressiveTypeInferencer(NodeTransformer):
    # TODO enforce all elements in a list to have the same type (length is not i.g. statically known!)

    # A stack of dictionaries for storing scoped knowledge of variable types
    scopes = [INITIAL_SCOPE]

    # Obtain the type of a variable name in the current scope
    def variable_type(self, name: str):
        for scope in reversed(self.scopes):
            if name in scope:
                return scope[name]
        raise TypeInferenceError(f"Variable {name} not initialized at access")

    def enter_scope(self):
        self.scopes.append({})

    def exit_scope(self):
        self.scopes.pop()

    def set_variable_type(self, name: str, typ: Type):
        if name in self.scopes[-1] and typ != self.scopes[-1][name]:
            raise TypeInferenceError(
                f"Type of variable {name} in local scope does not match inferred type {typ}"
            )
        self.scopes[-1][name] = typ

    def visit_ClassDef(self, node: ClassDef) -> ClassDef:
        class_record = RecordReader.extract(node)
        self.set_variable_type(node.name, RecordType(class_record))
        return node

    # TODO type inference for classDef

    def visit_Constant(self, node: Constant):
        tc = copy(node)
        assert type(node.value) not in [
            float,
            complex,
            type(...),
        ], "Float, complex numbers and ellipsis currently not supported"
        if tc.value is None:
            tc.typ = NoneType()
        else:
            tc.typ = RecordInstanceType(type(node.value).__name__)
        return tc

    def visit_Tuple(self, node: Tuple) -> TypedTuple:
        tt = copy(node)
        if not tt.elts:
            tt.typ = UnitType()
        else:
            tt.elts = [self.visit(e) for e in node.elts]
            tt.typ = TupleType([e.typ for e in tt.elts])
        return tt

    def visit_List(self, node: List) -> TypedList:
        tt = copy(node)
        tt.elts = [self.visit(e) for e in node.elts]
        tt.typ = ListType([e.typ for e in tt.elts])
        return tt

    def visit_Assign(self, node: Assign) -> TypedAssign:
        typed_ass = copy(node)
        typed_ass.value: TypedExpression = self.visit(node.value)
        # Make sure to first set the type of each target name so we can load it when visiting it
        for t in node.targets:
            if isinstance(t, Tuple):
                raise NotImplementedError("Type deconstruction not supported yet")
            self.set_variable_type(t.id, typed_ass.value.typ)
        typed_ass.targets = [self.visit(t) for t in node.targets]
        return typed_ass

    def visit_If(self, node: If) -> TypedIf:
        typed_if = copy(node)
        typed_if.test = self.visit(node.test)
        assert (
            typed_if.test.typ == BoolType
        ), "Branching condition must have boolean type"
        typed_if.body = [self.visit(s) for s in node.body]
        typed_if.orelse = [self.visit(s) for s in node.orelse]
        return typed_if

    def visit_While(self, node: While) -> TypedWhile:
        typed_while = copy(node)
        typed_while.test = self.visit(node.test)
        assert (
            typed_while.test.typ == BoolType
        ), "Branching condition must have boolean type"
        typed_while.body = [self.visit(s) for s in node.body]
        typed_while.orelse = [self.visit(s) for s in node.orelse]
        return typed_while

    def visit_For(self, node: For) -> TypedFor:
        typed_for = copy(node)
        typed_for.iter = self.visit(node.iter)
        if isinstance(node.target, Tuple):
            raise NotImplementedError(
                "Type deconstruction in for loops is not supported yet"
            )
        vartyp = None
        itertyp = typed_for.iter.typ
        if isinstance(itertyp, TupleType):
            vartyp = itertyp.typs[0]
            assert all(
                itertyp.typs[0] == t for t in typed_for.iter.typ.typs
            ), "Iterating through a tuple requires the same type for each element"
        elif isinstance(itertyp, ListType):
            vartyp = itertyp.typ
        else:
            raise NotImplementedError(
                "Type inference for non-list objects is not supported"
            )
        self.set_variable_type(node.target.id, vartyp)
        typed_for.target = self.visit(node.target)
        typed_for.body = [self.visit(s) for s in node.body]
        typed_for.orelse = [self.visit(s) for s in node.orelse]
        return typed_for

    def visit_Name(self, node: Name) -> TypedName:
        tn = copy(node)
        # Make sure that the rhs of an assign is evaluated first
        tn.typ = self.variable_type(node.id)
        return tn

    def visit_Compare(self, node: Compare) -> TypedCompare:
        typed_cmp = copy(node)
        typed_cmp.left = self.visit(node.left)
        typed_cmp.comparators = [self.visit(s) for s in node.comparators]
        typed_cmp.typ = BoolType
        assert all(
            typed_cmp.left.typ == c.typ for c in typed_cmp.comparators
        ), "Not all compared expressions have the same type"
        return typed_cmp

    def visit_arg(self, node: arg) -> typedarg:
        ta = copy(node)
        ta.typ = type_from_annotation(node.annotation)
        self.set_variable_type(ta.arg, ta.typ)
        return ta

    def visit_arguments(self, node: arguments) -> typedarguments:
        if node.kw_defaults or node.kwarg or node.kwonlyargs or node.defaults:
            raise NotImplementedError(
                "Keyword arguments and defaults not supported yet"
            )
        ta = copy(node)
        ta.args = [self.visit(a) for a in node.args]
        return ta

    def visit_FunctionDef(self, node: FunctionDef) -> TypedFunctionDef:
        tfd = copy(node)
        self.enter_scope()
        tfd.args = self.visit(node.args)
        tfd.typ = FunctionType(
            [t.typ for t in tfd.args.args],
            type_from_annotation(tfd.returns),
        )
        # We need the function type inside for recursion
        self.set_variable_type(node.name, tfd.typ)
        tfd.body = [self.visit(s) for s in node.body]
        # Check that return type and annotated return type match
        if not isinstance(node.body[-1], Return):
            assert (
                tfd.typ.rettyp == NoneType
            ), f"Function '{node.name}' has no return statement but is supposed to return not-None value"
        else:
            assert (
                tfd.typ.rettyp == tfd.body[-1].typ
            ), f"Function '{node.name}' annotated return type does not match actual return type"
        self.exit_scope()
        # We need the function type outside for usage
        self.set_variable_type(node.name, tfd.typ)
        return tfd

    def visit_Module(self, node: Module) -> TypedModule:
        self.enter_scope()
        tm = copy(node)
        tm.body = [self.visit(n) for n in node.body]
        self.exit_scope()
        return tm

    def visit_Expr(self, node: Expr) -> TypedExpr:
        tn = copy(node)
        tn.value = self.visit(node.value)
        return tn

    def visit_BinOp(self, node: BinOp) -> TypedBinOp:
        tb = copy(node)
        tb.left = self.visit(node.left)
        tb.right = self.visit(node.right)
        assert (
            tb.left.typ == tb.right.typ
        ), "Inputs to a binary operation need to have the same type"
        tb.typ = tb.left.typ
        return tb

    def visit_UnaryOp(self, node: UnaryOp) -> TypedUnaryOp:
        tu = copy(node)
        tu.operand = self.visit(node.operand)
        tu.typ = tu.operand.typ
        return tu

    def visit_Subscript(self, node: Subscript) -> TypedSubscript:
        ts = copy(node)
        assert isinstance(
            ts.slice, Index
        ), "Only single index slices are currently supported"
        # special case: Subscript of Union / Optional
        if isinstance(ts.value, Name) and ts.value.id in ["Union", "Optional"]:
            if ts.value.id == "Union":
                assert isinstance(
                    ts.slice.value, Tuple
                ), "Union must combine multiple classes"
                typs = self.visit(node.slice.value)
                eltyps = [e.typ for e in typs.elts]
                assert all(
                    isinstance(e, ClassType) for e in eltyps
                ), "Union must combine classes"
                ts.value = UnionType(eltyps)
                ts.typ = UnionType(eltyps)
                return ts
            if ts.value.id == "Optional":
                assert isinstance(
                    ts.slice.value, Name
                ), "Optional must have a single type as parameter"
                e = self.visit(node.slice.value)
                assert isinstance(
                    e.typ, ClassType
                ), "Optional must have a type as parameter"
                ts.value = OptionalType(e)
                ts.typ = OptionalType(e)
                return ts

        ts.value = self.visit(node.value)
        if isinstance(ts.value.typ, TupleType) or isinstance(ts.value.typ, ListType):
            if all(ts.value.typ.typs[0] == t for t in ts.value.typ.typs):
                ts.typ = ts.value.typ.typs[0]
            elif isinstance(ts.slice.value, Constant) and isinstance(
                ts.slice.value.value, int
            ):
                ts.typ = ts.value.typ.typs[ts.slice.value.value]
            else:
                raise TypeInferenceError(f"Could not infer type of subscript {node}")
        else:
            raise TypeInferenceError(f"Could not infer type of subscript {node}")
        return ts

    def visit_Call(self, node: Call) -> TypedCall:
        assert not node.keywords, "Keyword arguments are not supported yet"
        tc = copy(node)
        tc.func = self.visit(node.func)
        tc.args = [self.visit(a) for a in node.args]
        assert len(tc.args) == len(
            tc.func.typ.argtyps
        ), f"Signature of function {node} does not match number of arguments"
        assert all(
            a.typ == ap for a, ap in zip(tc.args, tc.func.typ.argtyps)
        ), f"Signature of function {node} does not match arguments"
        tc.typ = tc.func.typ.rettyp
        return tc

    def visit_Pass(self, node: Pass) -> TypedPass:
        tp = copy(node)
        return tp

    def visit_Return(self, node: Return) -> TypedReturn:
        tp = copy(node)
        tp.value = self.visit(node.value)
        tp.typ = tp.value.typ
        return tp

    def visit_Attribute(self, node: Attribute) -> TypedAttribute:
        tp = copy(node)
        tp.value = self.visit(node.value)
        owner = tp.value.typ
        assert isinstance(
            owner, RecordInstanceType
        ), "Accessing attribute of non-instance"
        # look up class type in local scope
        owner_typ = self.variable_type(owner.typ)
        assert isinstance(
            owner_typ, RecordType
        ), "Accessing attribute of instance of a non-class"
        tp.typ = None
        # TODO rewrite access to constructor
        for i, (attr_name, attr_type) in enumerate(owner_typ.record.fields):
            if attr_name == tp.attr:
                tp.typ = attr_type
                tp.pos = i
        if tp.typ is None:
            assert (
                tp.attr in tp.value.typ.record.fields
            ), "Accessing undefined attribute of class-type"
        return tp

    def generic_visit(self, node: AST) -> TypedAST:
        raise NotImplementedError(
            f"Cannot infer type of non-implemented node {node.__class__}"
        )


def typed_ast(ast: AST):
    return AggressiveTypeInferencer().visit(ast)
